package server

import (
	"fmt"
	"log"
	"net"
	"time"

	"github.com/gofiber/fiber/v2"

	"git.pbiernat.io/egommerce/api-entities/http"
	"git.pbiernat.io/egommerce/catalog-service/app"
	cnf "git.pbiernat.io/egommerce/catalog-service/config"
)

type (
	Server struct {
		*fiber.App
		cnf *cnf.ServerConfig

		addr string // e.g. "127.0.0.1:80"

		ShutdownFn
	}
	HeaderRequestID struct {
		RequestID string `reqHeader:"x-request-id"`
	}
	ShutdownFn func() error
)

func New(conf *cnf.Config) *Server {
	return &Server{
		App: fiber.New(fiber.Config{
			AppName:      conf.ID,
			ServerHeader: conf.Name + ":" + conf.ID,
			ReadTimeout:  conf.ReadTimeout * time.Millisecond,
			WriteTimeout: conf.WriteTimeout * time.Millisecond,
			IdleTimeout:  conf.IdleTimeout * time.Millisecond,
		}),
		cnf:  conf,
		addr: conf.NetAddr,
	}
}

func (s *Server) Init() {
	// svr.Base.ShutdownFn = svr.Shutdown()

	SetupMiddleware(s)
	SetupRouter(s)
}

func (s *Server) Start(while chan struct{}) {
	// fmt.Println("[DOER] Started server at: " + s.addr)
	ln, _ := net.Listen("tcp", s.addr)
	// ln = tls.NewListener(ln, s.App.Server().TLSConfig)
	err := s.Listener(ln)
	if err != nil {
		log.Fatalf("Failed to start server: %s. Reason: %v\n", s.cnf.ID, err)
		close(while)
	}
	<-while

	// return err
}

func (s *Server) OnShutdown(a *app.App) {
	a.Logger.Log("Server %s is going down...", a.cnf.ID)

	fmt.Println("Unregistering...")
	a.Registry.Unregister()
	// a.clearMetadataCache()
	a.Eventbus.Close()
	a.Database.Close()
	a.Logger.Log("Gone.")
	a.Logger.Close()
}

func (s *Server) GetRequestID(c *fiber.Ctx) (string, error) {
	var hdr = new(HeaderRequestID)
	if err := c.ReqHeaderParser(hdr); err != nil {
		return "", err
	}

	return hdr.RequestID, nil
}

func (s *Server) Error(c *fiber.Ctx, code int, msg string) error {
	return c.Status(code).JSON(http.ErrorResponse{Error: msg})
}

// @CHECK: merge s.Config and s.Base.Config to display all config as one array/map
// func (s *Server) registerKVUpdater() { // @FIXME: merge duplication in server.go and worker.go
// 	go func() {
// 		ticker := time.NewTicker(time.Second * 10)
// 		for range ticker.C {
// 			config, _, err := s.Registry.KV().Get(s.cnf.KVNamespace, nil)
// 			if err != nil || config == nil {
// 				return
// 			}

// 			kvCnf := bytes.NewBuffer(config.Value)
// 			decoder := json.NewDecoder(kvCnf)
// 			if err := decoder.Decode(&s.cnf); err != nil {
// 				return
// 			}
// 		}
// 	}()
// }

// // func (s *Server) clearMetadataCache() {
// // 	ctx := context.Background()
// // 	key, address := s.getMetadataIPsKey(), s.cnf.Base.AppID

// // 	s.Cache.LRem(ctx, key, 0, address)
// // }

// // func (s *Server) getMetadataIPsKey() string {
// // 	return "internal__" + s.cnf.Name + "__ips"
// // }
